use snafu::Report;
use snafu::{ResultExt, Snafu};
use std::{sync::Arc, time::Duration};
use tracing::error;

use uuid::Uuid;

use crate::storage;
use crate::storage::jobs::AddError;

/// Manage background job execution and storage.
#[derive(Clone, Debug)]
pub struct JobManager {
    /// Database connection for job storage.
    db: Arc<storage::jobs::Jobs>,
}

/// Errors that occur when running a background job.
#[derive(Debug, Snafu)]
pub enum ExecuteError {
    #[snafu(display("Failed to add new background job."))]
    Add { source: AddError },
}

impl JobManager {
    /// Create a new job manager with database connection.
    pub fn new(db: Arc<storage::jobs::Jobs>) -> Self {
        Self { db }
    }

    /// Execute a task in the background and return job ID.
    pub async fn execute<F, Fut, T, E>(&self, task: F) -> Result<Uuid, ExecuteError>
    where
        F: FnOnce() -> Fut + Send + 'static,
        Fut: Future<Output = Result<T, E>> + Send,
        T: serde::Serialize + Send + Sync + 'static,
        E: snafu::Error + Send + Sync + 'static,
    {
        let job_id = Uuid::new_v4();
        self.db.as_ref().add(&job_id).await.context(AddSnafu)?;
        let db = self.db.clone();

        tokio::spawn(async move {
            match task().await {
                Ok(result) => success(&job_id, &result, &db).await,
                Err(error) => failure(&job_id, error, &db).await,
            }

            tokio::time::sleep(Duration::from_secs(300)).await;
        });

        Ok(job_id)
    }
}

/// Mark job as successful and store result.
async fn success<T>(job_id: &Uuid, result: &T, db: &storage::jobs::Jobs)
where
    T: serde::Serialize + Send + Sync + 'static,
{
    let data = match postcard::to_allocvec(result) {
        Ok(data) => data,
        Err(error) => {
            failure(job_id, error, db).await;
            return;
        }
    };

    if let Err(error) = db.finish(job_id, &data).await {
        failure(job_id, error, db).await;
    }
}

/// Mark job as failed and log error.
async fn failure<E>(job_id: &Uuid, error: E, db: &storage::jobs::Jobs)
where
    E: snafu::Error + Send + Sync + 'static,
{
    let error = Report::from_error(&error);
    error!("{job_id}: {error}");
    if let Err(error) = db.fail(job_id, &error.to_string()).await {
        error!(
            "Failed to save job {job_id} result: {}",
            Report::from_error(&error)
        );
    }
}
