//! Query endpoint handlers and response types.

use std::{str::FromStr, sync::Arc};

use axum::{
    Json,
    extract::{Path, Query, State},
    http::StatusCode,
};
use serde::{Deserialize, Serialize};
use snafu::{ResultExt, Snafu, ensure};
use utoipa::ToSchema;
use uuid::Uuid;

use super::{TAG, error::HttpStatus, jobs, state::AppState};
use crate::{
    http_error, query,
    storage::{self, queries::DocumentMatch},
};

/// Maximum allowed limit for query results.
const MAX_LIMIT: usize = 10;

/// Errors that occur during query processing.
#[derive(Debug, Snafu)]
pub enum QueryError {
    #[snafu(display("'limit' query parameter must be a positive integer <= {MAX_LIMIT}."))]
    Limit,
    #[snafu(display("Failed to run query."))]
    QueryExecute { source: jobs::ExecuteError },
}

impl HttpStatus for QueryError {
    fn status_code(&self) -> StatusCode {
        match self {
            QueryError::Limit => StatusCode::BAD_REQUEST,
            QueryError::QueryExecute { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}

http_error!(QueryError);

/// Query parameters for search requests.
#[derive(Deserialize)]
pub struct QueryParams {
    /// Maximum number of results to return.
    pub limit: Option<usize>,
}

/// Response format for successful query requests.
#[derive(Debug, Serialize, ToSchema)]
pub struct QueryResponse {
    pub status: String,
    /// List of matching document chunks.
    pub results: Vec<DocumentResult>,
}

/// A single document search result.
#[derive(Debug, Serialize, ToSchema)]
pub struct DocumentResult {
    /// Calibre book ID containing this text.
    pub book_id: i64,
    /// Text content of the matching chunk.
    pub text_chunk: String,
    /// Similarity score between 0.0 and 1.0.
    pub similarity: f64,
}

impl From<DocumentMatch> for DocumentResult {
    fn from(doc: DocumentMatch) -> Self {
        Self {
            book_id: doc.book_id,
            text_chunk: doc.text_chunk,
            similarity: doc.similarity,
        }
    }
}

/// Response format for successful query requests.
#[derive(Debug, Serialize, ToSchema)]
pub struct QueryStartResponse {
    ///.Job id.
    pub id: String,
}

/// Execute a semantic search query against the document database and return the job id.
#[utoipa::path(
    post,
    path = "/query",
    tag = TAG,
    params(
        ("limit" = Option<usize>, Query, description = "Maximum number of results")
    ),
    request_body = String,
    responses(
        (status = 202, body = QueryStartResponse),
        (status = 400, description = "Wrong parameter."), 
        (status = 500, description = "Failure to query database.")
    )
)]
pub async fn start(
    Query(params): Query<QueryParams>,
    State(state): State<Arc<AppState>>,
    body: String,
) -> Result<Json<QueryStartResponse>, QueryError> {
    let limit = params.limit.unwrap_or(5);
    ensure!(limit <= MAX_LIMIT, LimitSnafu);

    let jobs = state.jobs.clone();
    let id = jobs
        .execute(move || async move {
            query::ask(
                &body,
                &state.db.queries,
                &state.tokenizer,
                &state.embedder,
                &state.reranker,
                state.chunk_size,
                limit,
            )
            .await
        })
        .await
        .context(QueryExecuteSnafu)?;

    Ok(Json(QueryStartResponse { id: id.to_string() }))
}

/// Errors that occur during query retrieval.
#[derive(Debug, Snafu)]
pub enum RetrieveError {
    #[snafu(display("Failed to retrieve job from database."))]
    Db {
        source: storage::jobs::RetrieveError,
    },
    #[snafu(display("No job with id {id}."))]
    NoJob { id: Uuid },
    #[snafu(display("'{id}' is not a valid v4 UUID."))]
    Uuid { source: uuid::Error, id: String },
}

impl HttpStatus for RetrieveError {
    fn status_code(&self) -> StatusCode {
        match self {
            RetrieveError::Db { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            RetrieveError::NoJob { id: _ } => StatusCode::NOT_FOUND,
            RetrieveError::Uuid { source: _, id: _ } => StatusCode::BAD_REQUEST,
        }
    }
}

http_error!(RetrieveError);

/// Get results for a completed query job.
#[utoipa::path(
    get,
    path = "/query/{id}",
    tag = TAG,
    responses(
        (status = OK, body = QueryResponse),
        (status = 404, description = "No job with the requested id."), 
        (status = 500, description = "Failure to query database.")
    )
)]
pub async fn result(
    Path(id): Path<String>,
    State(state): State<Arc<AppState>>,
) -> Result<Json<QueryResponse>, RetrieveError> {
    let id = Uuid::from_str(&id).context(UuidSnafu { id })?;
    match state.db.jobs.retrieve(&id).await.context(DbSnafu)? {
        Some((status, results)) => Ok(Json(QueryResponse {
            status,
            results: results.into_iter().map(Into::into).collect(),
        })),
        None => NoJobSnafu { id }.fail(),
    }
}
