use std::io::{self, Cursor, Read};

use scraper::Html;
use snafu::{ResultExt, Snafu};
use zip::result::ZipError;

use super::extractor::{EpubSnafu, ExtractionError, Extractor, extension};

/// EPUB text extractor.
pub struct Epub;

/// Errors that occur during EPUB text extraction.
#[derive(Debug, Snafu)]
pub enum EpubExtractionError {
    #[snafu(display("Failed to open epub as zip archive."))]
    OpenZip { source: ZipError },
    #[snafu(display("Failed to open file inside epub archive."))]
    ZipFileOpen { source: ZipError },
    #[snafu(display("Failed to read file inside epub archive."))]
    ZipFileRead { source: io::Error },
}

impl Extractor for Epub {
    fn extract(input: &[u8]) -> Result<String, ExtractionError> {
        let mut texts = Vec::new();
        let input = Cursor::new(input);
        let mut archive = zip::ZipArchive::new(input)
            .context(OpenZipSnafu)
            .context(EpubSnafu)?;
        for i in 0..archive.len() {
            let mut file = archive
                .by_index(i)
                .context(ZipFileOpenSnafu)
                .context(EpubSnafu)?;
            let extension = extension(file.enclosed_name().unwrap_or_default());
            if extension == "xhtml" || extension == "html" || extension == "htm" {
                let mut entry = String::new();
                file.read_to_string(&mut entry)
                    .context(ZipFileReadSnafu)
                    .context(EpubSnafu)?;

                let document = Html::parse_document(&entry);
                let text: String = document.root_element().text().collect::<Vec<_>>().join(" ");
                texts.push(text);
            }
        }

        Ok(texts.join(" "))
    }
}
