use std::{io::Read, path::Path};

use snafu::{ResultExt, Snafu};

use super::{
    epub::{Epub, EpubExtractionError},
    pdf::{Pdf, PdfExtractionError},
};

/// Errors that occur during text extraction from documents.
#[derive(Debug, Snafu)]
pub enum ExtractionError {
    #[snafu(display("Failed to read input for extraction."))]
    Read { source: std::io::Error },
    #[snafu(display("Failed to extract pdf."), visibility(pub))]
    Pdf { source: PdfExtractionError },
    #[snafu(display("Failed to extract epub."), visibility(pub))]
    Epub { source: EpubExtractionError },
}

/// Extract text content from document bytes.
pub trait Extractor {
    /// Extract text from document bytes.
    fn extract(input: &[u8]) -> Result<String, ExtractionError>;
}

/// Get file extension from path as lowercase string.
pub fn extension<P: AsRef<Path>>(path: P) -> String {
    let path = path.as_ref();
    path.extension()
        .unwrap_or_default()
        .to_string_lossy()
        .into_owned()
}

/// Extract text from input stream based on file extension.
pub fn extract<R: Read>(mut input: R, extension: &str) -> Result<Option<String>, ExtractionError> {
    let mut buffer = Vec::new();
    input.read_to_end(&mut buffer).context(ReadSnafu)?;

    Ok(match extension {
        "epub" => Some(Epub::extract(&buffer)?),
        "pdf" => Some(Pdf::extract(&buffer)?),
        _ => None,
    })
}
