//! Database storage and retrieval for document embeddings.

use std::sync::Arc;

use jobs::Jobs;
use queries::Queries;
use snafu::{ResultExt, Snafu};
use sqlx::postgres::PgPoolOptions;

pub mod jobs;
pub mod queries;

/// PostgreSQL database connection pool and operations.
#[derive(Debug, Clone)]
pub struct Postgres {
    pub jobs: Jobs,
    pub queries: Queries,
}

/// Error when connecting to database.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to connect to database."))]
pub struct ConnectionError {
    source: sqlx::Error,
}

impl Postgres {
    /// Create a new database connection pool.
    pub async fn connect(db_url: &str, max_connections: u32) -> Result<Self, ConnectionError> {
        let pool = Arc::new(
            PgPoolOptions::new()
                .max_connections(max_connections)
                .connect(db_url)
                .await
                .context(ConnectionSnafu)?,
        );

        let jobs = Jobs::new(pool.clone());
        let queries = Queries::new(pool.clone());

        Ok(Self { jobs, queries })
    }
}
