use std::sync::Arc;

use serde::{Deserialize, Serialize};
use snafu::{ResultExt, Snafu};
use sqlx::{PgPool, Row, Type};
use uuid::Uuid;

use super::queries::DocumentMatch;

/// Database operations for background job management.
#[derive(Debug, Clone)]
pub struct Jobs {
    /// Connection pool for database operations.
    pool: Arc<PgPool>,
}

/// Status of a background job.
#[derive(Debug, Clone, Serialize, Deserialize, Type)]
#[sqlx(type_name = "job_status", rename_all = "lowercase")]
pub enum JobStatus {
    Running,
    Completed,
    Failed,
}

impl std::fmt::Display for JobStatus {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            JobStatus::Running => write!(f, "running"),
            JobStatus::Completed => write!(f, "completed"),
            JobStatus::Failed => write!(f, "failed"),
        }
    }
}

impl std::str::FromStr for JobStatus {
    type Err = String;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "running" => Ok(JobStatus::Running),
            "completed" => Ok(JobStatus::Completed),
            "failed" => Ok(JobStatus::Failed),
            _ => Err(format!("Invalid job status: {}", s)),
        }
    }
}

/// Error adding a new job.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to add background job."))]
pub struct AddError {
    source: sqlx::Error,
}

/// Error when adding a result to a job.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to add result to background job."))]
pub struct ResultError {
    source: sqlx::Error,
}

/// Error when adding a result to a job.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to add failure to background job."))]
pub struct FailureError {
    source: sqlx::Error,
}

/// Errors that occur durin job retrieval.
#[derive(Debug, Snafu)]
pub enum RetrieveError {
    #[snafu(display("Failed to retrieve background job data"))]
    Retrieve { source: sqlx::Error },
    #[snafu(display("Failed to run query."))]
    Deserialize { source: postcard::Error },
}

impl Jobs {
    /// Create a new jobs handler with database pool.
    pub fn new(pool: Arc<PgPool>) -> Self {
        Self { pool }
    }

    /// Add a new job to the database and mark it as running.
    pub async fn add(&self, id: &Uuid) -> Result<(), AddError> {
        sqlx::query("INSERT INTO jobs (id, expires_at) VALUES ($1, NOW() + INTERVAL '24 hours')")
            .bind(id)
            .execute(self.pool.as_ref())
            .await
            .context(AddSnafu)?;

        Ok(())
    }

    /// Add a result to a job and mark it as finished.
    pub async fn finish(&self, id: &Uuid, data: &[u8]) -> Result<(), ResultError> {
        sqlx::query("UPDATE jobs SET status = 'completed', result = $1 WHERE id = $2")
            .bind(data)
            .bind(id)
            .execute(self.pool.as_ref())
            .await
            .context(ResultSnafu)?;

        Ok(())
    }

    /// Add an error to a job and mark it as failed.
    pub async fn fail(&self, id: &Uuid, error: &str) -> Result<(), FailureError> {
        sqlx::query("UPDATE jobs SET status = 'failed', error = $1 WHERE id = $2")
            .bind(error)
            .bind(id)
            .execute(self.pool.as_ref())
            .await
            .context(FailureSnafu)?;

        Ok(())
    }

    /// Retrieve job status and results by ID.
    pub async fn retrieve(
        &self,
        id: &Uuid,
    ) -> Result<Option<(String, Vec<DocumentMatch>)>, RetrieveError> {
        let row = sqlx::query("SELECT status, result FROM jobs WHERE id = $1")
            .bind(id)
            .fetch_optional(self.pool.as_ref())
            .await
            .context(RetrieveSnafu)?;

        match row {
            Some(row) => {
                let status: JobStatus = row.get("status");
                let status = status.to_string();
                match row.try_get("result") {
                    Ok(data) => {
                        let documents: Vec<DocumentMatch> =
                            postcard::from_bytes(data).context(DeserializeSnafu)?;
                        Ok(Some((status, documents)))
                    }
                    Err(_) => Ok(Some((status, Vec::new()))),
                }
            }
            None => Ok(None),
        }
    }
}
