//! Text embedding and reranking generation using ONNX models.

use std::{
    fs::File,
    io::{self, Read},
    path::Path,
    sync::Arc,
};

use rayon::prelude::*;
use snafu::{ResultExt, Snafu};
use tract_onnx::prelude::*;

use crate::{
    storage::queries::DocumentMatch,
    text_encoder::tract_ndarray::ShapeError,
    tokenize::{self, Encoding, Tokenizer},
};

/// Compiled ONNX model type alias.
type Model = SimplePlan<
    TypedFact,
    Box<dyn TypedOp>,
    tract_onnx::prelude::Graph<TypedFact, Box<dyn TypedOp>>,
>;

/// Embedding vector with optional offset and original text.
pub type Embeddings = (Vec<f32>, Option<usize>, String);

/// ONNX-based text embedder for generating vector representations.
#[derive(Debug, Clone)]
pub struct TextEncoder {
    /// Compiled ONNX model for inference.
    model: Model,
}

/// Errors when loading text encoderfrom file.
#[derive(Debug, Snafu)]
pub enum NewFromFileError {
    #[snafu(display("Failed to open model file."))]
    OpenModel { source: io::Error },
    #[snafu(display("Failed to instantiate model file."))]
    Instantiate { source: NewError },
}

/// Errors when creating text encoder from model data.
#[derive(Debug, Snafu)]
pub enum NewError {
    #[snafu(display("Failed to load model file."))]
    Model { source: TractError },
    #[snafu(display("Failed to optimize model."))]
    Optimization { source: TractError },
    #[snafu(display("Failed to convert model into runnable."))]
    Runnable { source: TractError },
}

/// Error when converting input to tensor format.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to convert inmput to i64 arc tensor."))]
pub struct ToArcTensorError {
    source: ShapeError,
}

/// Errors that occur during embedding generation.
#[derive(Debug, Snafu)]
pub enum EmbedError {
    #[snafu(display("Failed to run model."))]
    Run { source: TractError },
    #[snafu(display("Failed to transform results."))]
    Transform { source: TractError },
    #[snafu(display("Failed to pool results."))]
    Pooling,
    #[snafu(display("Failed to convert input."))]
    EmbedConversion { source: ToArcTensorError },
}

/// Errors that occur during reranking.
#[derive(Debug, Snafu)]
pub enum RerankError {
    #[snafu(display("Failed to tokenize query-document pair."))]
    Tokenize { source: tokenize::EncodeError },
    #[snafu(display("Failed to score pair."))]
    Score { source: ScoreError },
}

/// Errors that occur during embedding generation.
#[derive(Debug, Snafu)]
pub enum ScoreError {
    #[snafu(display("Failed to convert input."))]
    RerankConversion { source: ToArcTensorError },
    #[snafu(display("Failed to run model."))]
    RunRerank { source: TractError },
    #[snafu(display("Failed to transform results."))]
    TransformRerank { source: TractError },
}

impl TextEncoder {
    /// Create a text encoder from an ONNX model file.
    pub fn from_file<P: AsRef<Path>>(path: P) -> Result<Self, NewFromFileError> {
        let input = File::open(path).context(OpenModelSnafu)?;
        Self::new(&input).context(InstantiateSnafu)
    }

    /// Create a text encoder from ONNX model data.
    pub fn new<R: Read>(mut input: R) -> Result<Self, NewError> {
        let model = tract_onnx::onnx()
            .model_for_read(&mut input)
            .context(ModelSnafu)?
            .into_optimized()
            .context(OptimizationSnafu)?
            .into_runnable()
            .context(RunnableSnafu)?;

        Ok(Self { model })
    }

    /// Convert token IDs to i64 tensor format for model input.
    fn to_i64_arc_tensor(
        seq_length: usize,
        input: &[u32],
    ) -> Result<Arc<Tensor>, ToArcTensorError> {
        let input = Self::to_i64(input);
        let tensors = tract_ndarray::Array2::from_shape_vec((1, seq_length), input)
            .context(ToArcTensorSnafu)?
            .into_arc_tensor();

        Ok(tensors)
    }

    /// Convert u32 token IDs to i64 for model compatibility.
    fn to_i64(input: &[u32]) -> Vec<i64> {
        input.iter().map(|&id| id as i64).collect::<Vec<i64>>()
    }

    /// Generate embeddings for multiple text chunks in parallel.
    pub fn embed_multiple(&self, encodings: Vec<Encoding>) -> Result<Vec<Embeddings>, EmbedError> {
        encodings.into_par_iter().map(|e| self.embed(e)).collect()
    }

    /// Generate embedding vector for a single text chunk.
    pub fn embed(&self, encoding: Encoding) -> Result<Embeddings, EmbedError> {
        let seq_length = encoding.0.get_ids().len();

        let input_ids = Self::to_i64_arc_tensor(seq_length, encoding.0.get_ids())
            .context(EmbedConversionSnafu)?;
        let attention_mask = Self::to_i64_arc_tensor(seq_length, encoding.0.get_attention_mask())
            .context(EmbedConversionSnafu)?;
        let token_type_ids = tract_ndarray::Array2::<i64>::zeros((1, seq_length)).into_arc_tensor();

        let outputs = self
            .model
            .run(tvec!(
                TValue::Const(input_ids),
                TValue::Const(attention_mask),
                TValue::Const(token_type_ids),
            ))
            .context(RunSnafu)?;

        let embeddings = outputs[0]
            .to_array_view::<f32>()
            .context(TransformSnafu)?
            .into_owned();
        let pooled = embeddings
            .mean_axis(tract_ndarray::Axis(1))
            .ok_or(PoolingSnafu.build())?;

        // let embeddings = embeddings.into_raw_vec_and_offset();
        let embeddings = pooled.into_raw_vec_and_offset();

        Ok((embeddings.0, embeddings.1, encoding.1))
    }

    /// Rerank documents using cross-encoder model for better relevance scoring.
    pub fn rerank(
        &self,
        query: &str,
        documents: Vec<DocumentMatch>,
        tokenizer: &Tokenizer,
        limit: usize,
    ) -> Result<Vec<DocumentMatch>, RerankError> {
        let mut scored_docs = Vec::new();

        for doc in documents {
            let pair_text = format!("{} [SEP] {}", query, doc.text_chunk);

            let encodings = tokenizer.encode(&pair_text, 512).context(TokenizeSnafu)?;
            let encoding = &encodings[0].0;

            let score = self.score_pair(encoding).context(ScoreSnafu)?;

            let mut updated_doc = doc;
            updated_doc.similarity = score;
            scored_docs.push(updated_doc);
        }

        scored_docs.sort_by(|a, b| {
            b.similarity
                .partial_cmp(&a.similarity)
                .unwrap_or(std::cmp::Ordering::Equal)
        });
        Ok(scored_docs.into_iter().take(limit).collect())
    }

    /// Score a query-document pair using the reranking model.
    fn score_pair(&self, encoding: &tokenizers::Encoding) -> Result<f64, ScoreError> {
        let seq_length = encoding.get_ids().len();

        let input_ids = Self::to_i64_arc_tensor(seq_length, encoding.get_ids())
            .context(RerankConversionSnafu)?;
        let attention_mask = Self::to_i64_arc_tensor(seq_length, encoding.get_attention_mask())
            .context(RerankConversionSnafu)?;

        let outputs = self
            .model
            .run(tvec!(
                TValue::Const(input_ids),
                TValue::Const(attention_mask),
            ))
            .context(RunRerankSnafu)?;

        // Reranking model outputs a single relevance score
        let scores = outputs[0]
            .to_array_view::<f32>()
            .context(TransformRerankSnafu)?;
        let score = scores[[0, 0]] as f64;

        Ok(score)
    }
}
